@@ -85,11 +85,13 @@ let stack_delta = function
 	| ALocalVar -> -1
 	| ALocalAssign -> -2
 	| AReturn -> -1
+	| AFSCommand2 -> 0
 	| AGetURL2 _ -> -2
 	| ADeleteObj | AInstanceOf | ACast -> -1
 	| AExtends | AImplements -> -2
 	| AEnum2 | ATrace | AThrow -> -1
-	| AIncrement | ADecrement | AChr | AOrd | ARandom | ADelete | AGetTimer | ATypeOf | ATargetPath -> 0
+	| AGetTimer -> 1
+	| AIncrement | ADecrement | AChr | AOrd | ARandom | ADelete | ATypeOf | ATargetPath -> 0
 	| AObjCall | ACall | ANewMethod -> assert false
 	| op -> failwith ("Unknown stack delta for " ^ (ActionScript.action_string (fun _ -> "") 0 op))
 
@@ -300,8 +302,8 @@ let rec used_in_block curblock vname e =
 			List.exists vloop vl
 		| ECall (v,vl) ->
 			List.exists vloop (v :: vl)
-		| ENew (_,vl) ->
-			List.exists vloop vl
+		| ENew (v,vl) ->
+			vloop v || List.exists vloop vl
 		| EUnop (_,_,v) ->
 			vloop v
 		| EQuestion (v,v1,v2) ->
@@ -690,6 +692,12 @@ and generate_call ?(newcall=false) ctx v vl =
 	| EConst (Ident "targetPath") , [v] ->
 		generate_val ctx v;
 		write ctx ATargetPath
+	| EConst (Ident "FSCommand2") , l ->
+		List.iter (generate_val ctx) (List.rev l);		
+		let nargs = List.length l in
+		push ctx [VInt nargs];
+		write ctx AFSCommand2;
+		ctx.stack <- ctx.stack - (nargs - 1)
 	| EConst (Ident "fscommand") , [v] ->
 		push ctx [VStr "FSCommand:"];
 		generate_val ctx v;
@@ -824,7 +832,7 @@ and generate_val ?(retval=true) ctx (v,p) =
 		setvar ~retval:(retval && flag = Prefix) ctx k
 
 let generate_local_var ctx (vname,_,vinit) =
-	if used_in_block false vname ctx.cur_block then begin
+	if used_in_block false vname ctx.cur_block || ctx.reg_count >= 250 then begin
 		push ctx [VStr vname];
 		Hashtbl.add ctx.locals vname { reg = 0; sp = ctx.stack };
 		match vinit with
@@ -1508,6 +1516,14 @@ let make_header s =
 	with
 		_ -> raise (Arg.Bad "Invalid header format")
 
+let rec trim f =
+	let l = String.length f in
+	if l = 0 then
+		""
+	else match f.[l - 1] with
+		| '\r' | '\n' -> trim (String.sub f 0 (l - 1))
+		| _ -> f
+
 let exclude_file f =
 	let lines = (try
 		let ch = open_in (Plugin.find_file f) in
@@ -1518,6 +1534,7 @@ let exclude_file f =
 		String.nsplit f ";"
 	) in
 	List.iter (fun f ->
+		let f = trim f in
 		if f <> "" then Hashtbl.replace excludes f ()
 	) lines
 
